{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# Recurrent networks\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "There are two different approaches to using BackPACK with RNNs.\n\n1. `Custom RNN with BackPACK custom modules`:\n   Build your RNN with custom modules provided by BackPACK\n   without overwriting the forward pass. This approach is useful if you want to\n   understand how BackPACK handles RNNs, or if you think building a container\n   module that implicitly defines the forward pass is more elegant than coding up\n   a forward pass.\n\n2. `RNN with BackPACK's converter`:\n   Automatically convert your model into a BackPACK-compatible architecture.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>RNNs are still an experimental feature. Always double-check your\n   results, as done in this example! Open an issue if you encounter a bug to help\n   us improve the support.\n\n   Not all extensions support RNNs (yet). Please create a feature request in the\n   repository if the extension you need is not supported.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's get the imports out of the way.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "from torch import (\n    allclose,\n    cat,\n    device,\n    int32,\n    linspace,\n    manual_seed,\n    nn,\n    randint,\n    zeros_like,\n)\n\nfrom backpack import backpack, extend\nfrom backpack.custom_module.graph_utils import BackpackTracer\nfrom backpack.custom_module.permute import Permute\nfrom backpack.custom_module.reduce_tuple import ReduceTuple\nfrom backpack.extensions import BatchGrad, DiagGGNExact\nfrom backpack.utils.examples import autograd_diag_ggn_exact\n\nmanual_seed(0)\nDEVICE = device(\"cpu\")  # Verification via autograd only works on CPU"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "For this demo, we will use the Tolstoi Char RNN from\n`DeepOBS <https://github.com/fsschneider/DeepOBS>`_.\nThis network is trained on Leo Tolstoi's War and Peace\nand learns to predict the next character.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class TolstoiCharRNN(nn.Module):\n    def __init__(self):\n        super().__init__()\n        self.batch_size = 8\n        self.hidden_dim = 64\n        self.num_layers = 2\n        self.seq_len = 15\n        self.vocab_size = 25\n\n        self.embedding = nn.Embedding(\n            num_embeddings=self.vocab_size, embedding_dim=self.hidden_dim\n        )\n        self.dropout = nn.Dropout(p=0.2)\n        self.lstm = nn.LSTM(\n            input_size=self.hidden_dim,\n            hidden_size=self.hidden_dim,\n            num_layers=self.num_layers,\n            dropout=0.36,\n            batch_first=True,\n        )\n        # deactivate redundant bias\n        self.lstm.bias_ih_l0.data = zeros_like(self.lstm.bias_ih_l0)\n        self.lstm.bias_ih_l1.data = zeros_like(self.lstm.bias_ih_l1)\n        self.lstm.bias_ih_l0.requires_grad = False\n        self.lstm.bias_ih_l1.requires_grad = False\n        self.dense = nn.Linear(\n            in_features=self.hidden_dim, out_features=self.vocab_size\n        )\n\n    def forward(self, x):\n        x = self.embedding(x)\n        x = self.dropout(x)\n        x, _ = self.lstm(x)  # last return values are hidden states\n        x = self.dropout(x)\n        output = self.dense(x)\n        output = output.permute(0, 2, 1)  # [N, T, C] \u2192 [N, C, T]\n        return output\n\n    def input_target_fn(self):\n        input = randint(0, self.vocab_size, (self.batch_size, self.seq_len))\n        # target is the input shifted by 1 in time axis\n        target = cat(\n            [\n                randint(0, self.vocab_size, (self.batch_size, 1)),\n                input[:, :-1],\n            ],\n            dim=1,\n        )\n        return input.to(DEVICE), target.to(DEVICE)\n\n    def loss_fn(self) -> nn.Module:\n        return nn.CrossEntropyLoss().to(DEVICE)\n\n\nmanual_seed(1)\ntolstoi_char_rnn = TolstoiCharRNN().to(DEVICE).eval()\nloss_function = extend(tolstoi_char_rnn.loss_fn())\nx, y = tolstoi_char_rnn.input_target_fn()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Note that instead of the real data set, we will feed synthetic data to the network for\nsimplicity. We also use the network in evaluation mode. This disables the\n:py:class:`Dropout <torch.nn.Dropout>` layers and allows double-checking our results\nvia :py:mod:`torch.autograd`.\n\n## Custom RNN with BackPACK custom modules\nSecond-order extensions only work if every node in the computation graph is an\n``nn`` module that can be extended by BackPACK. The above RNN\n:py:class:`TolstoiCharRNN<TolstoiCharRNN>` does not satisfy these conditions, because\nit has a multi-layer :py:class:`torch.nn.LSTM` and implicitly uses the\n:py:func:`getitem` (for unpacking) and :py:meth:`permute() <torch.Tensor.permute>`\nfunctions in the :py:meth:`forward() <torch.nn.Module.forward>` method.\n\nTo build RNN without overwriting the forward pass, BackPACK offers custom modules:\n\n1. :py:class:`ReduceTuple <backpack.custom_module.reduce_tuple.ReduceTuple>`\n\n2. :py:class:`Permute <backpack.custom_module.permute.Permute>`\n\nWith the above modules, we can build a simple RNN as a container that implicitly\ndefines the forward pass:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "manual_seed(1)  # same seed as used to initialize `tolstoi_char_rnn`\ntolstoi_char_rnn_custom = nn.Sequential(\n    nn.Embedding(tolstoi_char_rnn.vocab_size, tolstoi_char_rnn.hidden_dim),\n    nn.Dropout(p=0.2),\n    nn.LSTM(tolstoi_char_rnn.hidden_dim, tolstoi_char_rnn.hidden_dim, batch_first=True),\n    ReduceTuple(index=0),\n    nn.Dropout(p=0.36),\n    nn.LSTM(tolstoi_char_rnn.hidden_dim, tolstoi_char_rnn.hidden_dim, batch_first=True),\n    ReduceTuple(index=0),\n    nn.Dropout(p=0.2),\n    nn.Linear(tolstoi_char_rnn.hidden_dim, tolstoi_char_rnn.vocab_size),\n    Permute(0, 2, 1),\n)\ntolstoi_char_rnn_custom.eval().to(DEVICE)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's check that both models have the same forward pass.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "for name, p in tolstoi_char_rnn_custom.named_parameters():\n    if \"bias_ih_l\" in name:\n        # deactivate redundant bias\n        p.data = zeros_like(p.data)\n        p.requires_grad = False\n\nmatch = allclose(tolstoi_char_rnn_custom(x), tolstoi_char_rnn(x))\nprint(f\"Forward pass of custom model matches TolstoiCharRNN? {match}\")\n\nif not match:\n    raise AssertionError(\"Forward passes don't match.\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We can :py:func:`extend <backpack.extend>` our model and the loss function to\ncompute BackPACK extensions.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "tolstoi_char_rnn_custom = extend(tolstoi_char_rnn_custom)\nloss = loss_function(tolstoi_char_rnn_custom(x), y)\n\nwith backpack(BatchGrad(), DiagGGNExact()):\n    loss.backward()\n\nfor name, param in tolstoi_char_rnn_custom.named_parameters():\n    if param.requires_grad:\n        print(\n            name,\n            param.shape,\n            param.grad_batch.shape,\n            param.diag_ggn_exact.shape,\n        )"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Comparison of the GGN diagonal extension with :py:mod:`torch.autograd`:\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Computing the full GGN diagonal with PyTorch's built-in automatic differentiation\n   can be slow, depending on the number of parameters. To reduce run time, we only\n   compare some elements of the diagonal.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "trainable_params = [p for p in tolstoi_char_rnn_custom.parameters() if p.requires_grad]\n\ndiag_ggn_exact_vector = cat([p.diag_ggn_exact.flatten() for p in trainable_params])\n\nnum_params = sum(p.numel() for p in trainable_params)\nnum_to_compare = 10\nidx_to_compare = linspace(0, num_params - 1, num_to_compare, device=DEVICE, dtype=int32)\n\ndiag_ggn_exact_to_compare = autograd_diag_ggn_exact(\n    x, y, tolstoi_char_rnn_custom, loss_function, idx=idx_to_compare\n)\n\nprint(\"Do the exact GGN diagonals match?\")\nfor idx, element in zip(idx_to_compare, diag_ggn_exact_to_compare):\n    match = allclose(element, diag_ggn_exact_vector[idx])\n    print(\n        f\"Diagonal entry {idx:>8}: {match}:\"\n        + f\"\\t{element:.5e}, {diag_ggn_exact_vector[idx]:.5e}\"\n    )\n    if not match:\n        raise AssertionError(\"Exact GGN diagonals don't match!\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## RNN with BackPACK's converter\nIf you are not building an RNN through custom modules but for instance want to\ndirectly use the Tolstoi Char RNN, BackPACK offers a converter.\nIt analyzes the model and tries to turn it into a compatible architecture. The result\nis a :py:class:`torch.fx.GraphModule` that exclusively consists of modules.\n\nHere, we demonstrate the converter on the above Tolstoi Char RNN. Let's convert it\nwhile :py:func:`extend <backpack.extend>`-ing the model:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "# use BackPACK's converter to extend the model (turned off by default)\ntolstoi_char_rnn = extend(tolstoi_char_rnn, use_converter=True)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "To get an understanding what happened, we can inspect the model's graph with the\nfollowing helper function:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def print_table(module: nn.Module) -> None:\n    \"\"\"Prints a table of the module.\n\n    Args:\n        module: module to analyze\n    \"\"\"\n    graph = BackpackTracer().trace(module)\n    graph.print_tabular()\n\n\nprint_table(tolstoi_char_rnn)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Note that the computation graph fully consists of modules (indicated by\n``call_module`` in the first table column) such that BackPACK's hooks can\nsuccessfully backpropagate additional information for its second-order extensions\n(first-order extensions work, too).\n\nFirst, let's compare the forward pass with the custom module from the previous\nsection to make sure the converter worked fine:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "match = allclose(tolstoi_char_rnn_custom(x), tolstoi_char_rnn(x))\nprint(f\"Forward pass of extended TolstoiCharRNN matches custom model? {match}\")\n\nif not match:\n    raise AssertionError(\"Forward passes don't match.\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Now let's verify that second-order extensions (GGN diagonal) are working:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "output = tolstoi_char_rnn(x)\nloss = loss_function(output, y)\n\nwith backpack(DiagGGNExact()):\n    loss.backward()\n\nfor name, parameter in tolstoi_char_rnn.named_parameters():\n    if parameter.requires_grad:\n        print(f\"{name}'s diag_ggn_exact: {parameter.diag_ggn_exact.shape}\")\n\ndiag_ggn_exact_vector = cat(\n    [\n        p.diag_ggn_exact.flatten()\n        for p in tolstoi_char_rnn.parameters()\n        if p.requires_grad\n    ]\n)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Finally, we compare BackPACK's GGN diagonal with :py:mod:`torch.autograd`:\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Computing the full GGN diagonal with PyTorch's built-in automatic differentiation\n   can be slow, depending on the number of parameters. To reduce run time, we only\n   compare some elements of the diagonal.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "num_params = sum(p.numel() for p in tolstoi_char_rnn.parameters() if p.requires_grad)\nnum_to_compare = 10\nidx_to_compare = linspace(0, num_params - 1, num_to_compare, device=DEVICE, dtype=int32)\n\ndiag_ggn_exact_to_compare = autograd_diag_ggn_exact(\n    x, y, tolstoi_char_rnn, loss_function, idx=idx_to_compare\n)\n\nprint(\"Do the exact GGN diagonals match?\")\nfor idx, element in zip(idx_to_compare, diag_ggn_exact_to_compare):\n    match = allclose(element, diag_ggn_exact_vector[idx])\n    print(\n        f\"Diagonal entry {idx:>8}: {match}:\"\n        + f\"\\t{element:.5e}, {diag_ggn_exact_vector[idx]:.5e}\"\n    )\n    if not match:\n        raise AssertionError(\"Exact GGN diagonals don't match!\")"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}